import romt.crate


def test_crate_prefix_from_name() -> None:
    lower = romt.crate.PrefixStyle.LOWER
    mixed = romt.crate.PrefixStyle.MIXED
    assert romt.crate.crate_prefix_from_name("a", lower) == "1"
    assert romt.crate.crate_prefix_from_name("a", mixed) == "1"
    assert romt.crate.crate_prefix_from_name("ab", lower) == "2"
    assert romt.crate.crate_prefix_from_name("ab", mixed) == "2"
    assert romt.crate.crate_prefix_from_name("AbC", lower) == "3/a"
    assert romt.crate.crate_prefix_from_name("AbC", mixed) == "3/A"
    assert romt.crate.crate_prefix_from_name("AbCd", lower) == "ab/cd"
    assert romt.crate.crate_prefix_from_name("AbCd", mixed) == "Ab/Cd"



import filecmp
import hashlib
import json
import os
import shutil
import textwrap
import typing as T
from pathlib import Path

import git
import pytest
import toml

import romt.cli
import romt.crate


def path_append_suffix(path: Path, suffix: str) -> Path:
    return path.with_suffix(path.suffix + suffix)


def rmtree(path: Path) -> None:
    if path.is_dir():
        shutil.rmtree(path)


def mkdir_p(path: Path) -> None:
    if not path.is_dir():
        path.mkdir(parents=True)


def write_file_bytes(path: Path, data: bytes) -> None:
    mkdir_p(path.parent)
    path.write_bytes(data)


def write_file_text(path: Path, text: str) -> None:
    write_file_bytes(path, text.encode())


def append_file_line(path: Path, line: str) -> None:
    mkdir_p(path.parent)
    with open(path, "a") as f:
        f.write(line + "\n")


def walk_files(root: Path) -> T.Generator[Path, None, None]:
    for dirpath, dirnames, filenames in os.walk(root):
        for name in filenames:
            yield Path(dirpath) / name


def walk_files_rel(root: Path) -> T.Generator[str, None, None]:
    for p in walk_files(root):
        yield str(p.relative_to(root))


def assert_same_files(
    left_root: Path,
    left_files_rel: T.Iterable[str],
    right_root: Path,
    right_files_rel: T.Iterable[str],
) -> None:
    left = set(left_files_rel)
    right = set(right_files_rel)
    common = left & right
    left_only = left - right
    right_only = right - left

    match, mismatch, errors = filecmp.cmpfiles(
        left_root, right_root, common, shallow=False
    )

    print(f"{left_only=}")
    print(f"{right_only=}")
    print(f"{mismatch=}")
    print(f"{errors=}")
    assert not (left_only or right_only or mismatch or errors)


def create_repo(repo_path: Path) -> git.Repo:
    mkdir_p(repo_path)
    repo = git.Repo.init(str(repo_path))
    master = git.Reference(repo, "refs/heads/master")
    repo.head.set_reference(master)
    return repo


def repo_work_path(repo: git.Repo) -> Path:
    working_tree_dir = repo.working_tree_dir
    assert working_tree_dir is not None
    return Path(working_tree_dir)


def repo_add_config(repo: git.Repo) -> None:
    config_path = repo_work_path(repo) / "config.json"
    config = dict(
        dl="https://crates.io/api/v1/crates", api="https://crates.io"
    )
    write_file_text(config_path, json.dumps(config, indent=2))
    repo.index.add(str(config_path))
    repo.index.commit("add `config.json`")


def make_crate_entry(
    name: str, version: str, sha256sum: str
) -> T.Dict[str, T.Any]:
    return dict(
        name=name,
        vers=version,
        deps=[
            # {
            #     "name": "anyhow",
            #     "req": "^1.0.66",
            #     "features": [],
            #     "optional": false,
            #     "default_features": true,
            #     "target": null,
            #     "kind": "normal",
            # }
        ],
        cksum=sha256sum,
        features={},
        yanked=False,
    )


def add_crate(
    repo: git.Repo, crates_path: Path, name: str, version: str
) -> None:
    prefix = romt.crate.crate_prefix_from_name(
        name, romt.crate.PrefixStyle.LOWER
    )
    crate_path = crates_path / prefix / name / f"{name}-{version}.crate"
    crate_data = (crate_path.name + "\n").encode()
    write_file_bytes(crate_path, crate_data)

    work_path = repo_work_path(repo)
    entry_path = work_path / prefix / name
    sha256sum = hashlib.sha256(crate_data).hexdigest()
    entry = make_crate_entry(name, version, sha256sum)
    append_file_line(entry_path, json.dumps(entry))
    repo.index.add(str(entry_path))
    repo.index.commit(f"add `{crate_path.name}`")


CRATE_VERSIONS = [
    ("a", "0.1.0"),
    ("ab", "0.1.0"),
    ("abc", "0.1.0"),
    ("abcd", "0.1.0"),
    ("abcdefgh", "0.1.0"),
]


# Create `path` (containing `data`) and `path.sha256`, returning `sha256sum`.
def make_sha256_file_pair(path: Path, data: bytes) -> str:
    write_file_bytes(path, data)
    sha256sum = hashlib.sha256(data).hexdigest()
    write_file_text(path_append_suffix(path, ".sha256"), sha256sum + "\n")
    return sha256sum


# Create `path` and `path.sha256`, returning `sha256sum`.
def make_fake_sha256_file_pair(path: Path) -> str:
    data = (path.name + "\n").encode()
    return make_sha256_file_pair(path, data)


def make_dist_file(root: Path, url: str) -> str:
    prefix = "https://static.rust-lang.org/"
    assert url.startswith(prefix)
    path = root / url[len(prefix) :]
    return make_fake_sha256_file_pair(path)


def make_dist(root: Path, manifest: T.Any) -> None:
    if isinstance(manifest, dict):
        url = manifest.get("url")
        if url:
            manifest["hash"] = make_dist_file(root, url)
        xz_url = manifest.get("xz_url")
        if xz_url:
            manifest["xz_hash"] = make_dist_file(root, xz_url)
        for value in manifest.values():
            make_dist(root, value)
    elif isinstance(manifest, list):
        for value in manifest:
            make_dist(root, value)


@pytest.fixture
def tests_path(request: pytest.FixtureRequest) -> Path:
    return Path(request.path).parent


@pytest.fixture
def upstream_path(tests_path: Path) -> Path:
    path = Path("fake") / "upstream"
    rmtree(path)

    # Setup upstream crates.
    crates_path = path / "crates"
    mkdir_p(crates_path)
    repo_path = path / "git" / "crates.io-index"
    repo = create_repo(repo_path)
    repo_add_config(repo)
    for name, version in CRATE_VERSIONS:
        add_crate(repo, crates_path, name, version)

    # Setup upstream toolchain.
    manifest = toml.load(tests_path / "channel-rust-1.76.0.toml")
    make_dist(path, manifest)
    manifest_bytes = toml.dumps(manifest).encode()
    dist_path = path / "dist"
    for version in ["1.76.0", "stable"]:
        for p in [dist_path, dist_path / "2024-02-08"]:
            make_sha256_file_pair(
                p / f"channel-rust-{version}.toml", manifest_bytes
            )

    # Setup upstream `rustup`.
    rustup_path = path / "rustup"
    rustup_toml_path = rustup_path / "release-stable.toml"
    write_file_text(
        rustup_toml_path,
        textwrap.dedent(
            """\
            schema-version = "1"
            version = "1.26.0"
            """
        ),
    )

    target = "x86_64-unknown-linux-gnu"
    version = "1.26.0"
    make_fake_sha256_file_pair(
        rustup_path / "archive" / version / target / "rustup-init"
    )
    make_fake_sha256_file_pair(rustup_path / "dist" / target / "rustup-init")

    return path


@pytest.fixture
def inet_path() -> Path:
    inet_path = Path("fake") / "inet"
    rmtree(inet_path)
    mkdir_p(inet_path)
    return inet_path


@pytest.fixture
def offline_path() -> Path:
    offline_path = Path("fake") / "offline"
    rmtree(offline_path)
    mkdir_p(offline_path)
    return offline_path


def crate_must_run(root_path: Path, args: T.List[str]) -> None:
    crate_args = [
        "crate",
        "--crates",
        f"{root_path}/crates",
        "--index",
        f"{root_path}/git/crates.io-index",
    ]
    assert romt.cli.run(crate_args + args) == 0


def test_crates(
    upstream_path: Path, inet_path: Path, offline_path: Path
) -> None:
    upstream_crates_path = upstream_path / "crates"
    upstream_files_rel = set(walk_files_rel(upstream_crates_path))

    crates_template = "crates/{lowerprefix}/{crate}/{crate}-{version}.crate"
    archive_path = inet_path / "crates.tar.gz"
    inet_args = [
        "--crates-url",
        f"{upstream_path}/{crates_template}",
        "--index-url",
        f"{upstream_path}/git/crates.io-index",
        "--archive",
        f"{archive_path}",
    ]

    crate_must_run(inet_path, inet_args + ["init"])
    crate_must_run(inet_path, inet_args + ["export"])

    inet_crates_path = inet_path / "crates"
    inet_files_rel = set(walk_files_rel(inet_crates_path))
    inet_files_rel.remove("config.toml")

    assert_same_files(
        upstream_crates_path,
        upstream_files_rel,
        inet_crates_path,
        inet_files_rel,
    )

    offline_args = [
        "--crates-url",
        f"{inet_path}/{crates_template}",
        "--index-url",
        f"{inet_path}/git/crates.io-index",
        "--archive",
        f"{archive_path}",
    ]

    crate_must_run(offline_path, offline_args + ["init-import"])
    crate_must_run(offline_path, offline_args + ["import"])

    offline_crates_path = offline_path / "crates"
    offline_files_rel = set(walk_files_rel(offline_crates_path))
    offline_files_rel.remove("config.toml")

    assert_same_files(
        inet_crates_path,
        inet_files_rel,
        offline_crates_path,
        offline_files_rel,
    )


def toolchain_must_run(root_path: Path, args: T.List[str]) -> None:
    toolchain_args = [
        "toolchain",
        "--dest",
        f"{root_path}/dist",
    ]
    assert romt.cli.run(toolchain_args + args) == 0


toolchain_artifact_names_1_76_0 = set(
    """

    cargo-1.76.0-x86_64-unknown-linux-gnu.tar.xz
    cargo-1.76.0-x86_64-unknown-linux-gnu.tar.xz.sha256
    channel-rust-1.76.0.toml
    channel-rust-1.76.0.toml
    channel-rust-1.76.0.toml.sha256
    channel-rust-1.76.0.toml.sha256
    channel-rust-stable.toml
    channel-rust-stable.toml
    channel-rust-stable.toml.sha256
    channel-rust-stable.toml.sha256
    clippy-1.76.0-x86_64-unknown-linux-gnu.tar.xz
    clippy-1.76.0-x86_64-unknown-linux-gnu.tar.xz.sha256
    llvm-tools-1.76.0-x86_64-unknown-linux-gnu.tar.xz
    llvm-tools-1.76.0-x86_64-unknown-linux-gnu.tar.xz.sha256
    reproducible-artifacts-1.76.0-x86_64-unknown-linux-gnu.tar.xz
    reproducible-artifacts-1.76.0-x86_64-unknown-linux-gnu.tar.xz.sha256
    rls-1.76.0-x86_64-unknown-linux-gnu.tar.xz
    rls-1.76.0-x86_64-unknown-linux-gnu.tar.xz.sha256
    rust-1.76.0-x86_64-unknown-linux-gnu.tar.xz
    rust-1.76.0-x86_64-unknown-linux-gnu.tar.xz.sha256
    rust-analysis-1.76.0-x86_64-unknown-linux-gnu.tar.xz
    rust-analysis-1.76.0-x86_64-unknown-linux-gnu.tar.xz.sha256
    rust-analyzer-1.76.0-x86_64-unknown-linux-gnu.tar.xz
    rust-analyzer-1.76.0-x86_64-unknown-linux-gnu.tar.xz.sha256
    rust-docs-1.76.0-x86_64-unknown-linux-gnu.tar.xz
    rust-docs-1.76.0-x86_64-unknown-linux-gnu.tar.xz.sha256
    rust-src-1.76.0.tar.xz
    rust-src-1.76.0.tar.xz.sha256
    rust-std-1.76.0-x86_64-unknown-linux-gnu.tar.xz
    rust-std-1.76.0-x86_64-unknown-linux-gnu.tar.xz.sha256
    rustc-1.76.0-x86_64-unknown-linux-gnu.tar.xz
    rustc-1.76.0-x86_64-unknown-linux-gnu.tar.xz.sha256
    rustc-dev-1.76.0-x86_64-unknown-linux-gnu.tar.xz
    rustc-dev-1.76.0-x86_64-unknown-linux-gnu.tar.xz.sha256
    rustc-docs-1.76.0-x86_64-unknown-linux-gnu.tar.xz
    rustc-docs-1.76.0-x86_64-unknown-linux-gnu.tar.xz.sha256
    rustfmt-1.76.0-x86_64-unknown-linux-gnu.tar.xz
    rustfmt-1.76.0-x86_64-unknown-linux-gnu.tar.xz.sha256


""".split()
)


def rel_paths_with_base_names(
    rel_paths: T.Iterable[str], base_names: T.Iterable[str]
) -> T.Generator[str, None, None]:
    names = set(base_names)
    for p in rel_paths:
        if os.path.basename(p) in names:
            yield p


def test_toolchain(
    upstream_path: Path,
    inet_path: Path,
    offline_path: Path,
) -> None:
    upstream_dist_path = upstream_path / "dist"
    upstream_files_rel = set(walk_files_rel(upstream_dist_path))

    inet_dist_path = inet_path / "dist"
    archive_path = inet_path / "toolchain.tar.gz"
    inet_args = [
        "--url",
        str(upstream_dist_path),
        "--archive",
        f"{archive_path}",
        "--no-signature",
    ]

    artifact_names = set(toolchain_artifact_names_1_76_0)

    toolchain_must_run(
        inet_path,
        inet_args + ["-s", "1.76.0", "-t", "linux", "download", "pack"],
    )

    inet_files_rel = set(walk_files_rel(inet_dist_path))

    assert_same_files(
        upstream_dist_path,
        rel_paths_with_base_names(upstream_files_rel, artifact_names),
        inet_dist_path,
        inet_files_rel,
    )

    offline_dist_path = offline_path / "dist"
    offline_args = [
        "--archive",
        f"{archive_path}",
        "--no-signature",
    ]

    toolchain_must_run(
        offline_path,
        offline_args + ["unpack"],
    )

    offline_files_rel = set(walk_files_rel(offline_dist_path))

    assert_same_files(
        inet_dist_path,
        inet_files_rel,
        offline_dist_path,
        offline_files_rel,
    )

    # Now include a `--cross` for a single target.
    cross_target = "x86_64-unknown-linux-musl"
    toolchain_must_run(
        inet_path,
        inet_args
        + ["-s", "1.76.0", "-t", cross_target, "--cross", "download", "pack"],
    )

    artifact_names.update(
        [
            "rust-std-1.76.0-x86_64-unknown-linux-musl.tar.xz",
            "rust-std-1.76.0-x86_64-unknown-linux-musl.tar.xz.sha256",
        ]
    )

    inet_files_rel = set(walk_files_rel(inet_dist_path))

    assert_same_files(
        upstream_dist_path,
        rel_paths_with_base_names(upstream_files_rel, artifact_names),
        inet_dist_path,
        inet_files_rel,
    )

    toolchain_must_run(
        offline_path,
        offline_args + ["unpack"],
    )

    offline_files_rel = set(walk_files_rel(offline_dist_path))

    assert_same_files(
        inet_dist_path,
        inet_files_rel,
        offline_dist_path,
        offline_files_rel,
    )


def rustup_must_run(root_path: Path, args: T.List[str]) -> None:
    toolchain_args = [
        "rustup",
        "--dest",
        f"{root_path}/rustup",
    ]
    assert romt.cli.run(toolchain_args + args) == 0


def test_rustup(
    upstream_path: Path,
    inet_path: Path,
    offline_path: Path,
) -> None:
    upstream_rustup_path = upstream_path / "rustup"
    upstream_files_rel = set(walk_files_rel(upstream_rustup_path))

    inet_rustup_path = inet_path / "rustup"
    archive_path = inet_path / "rustup.tar.gz"
    inet_args = [
        "--url",
        str(upstream_rustup_path),
        "--archive",
        f"{archive_path}",
    ]

    rustup_must_run(
        inet_path,
        inet_args + ["-s", "stable", "-t", "linux", "download", "pack"],
    )

    inet_files_rel = set(walk_files_rel(inet_rustup_path))

    assert_same_files(
        upstream_rustup_path,
        upstream_files_rel,
        inet_rustup_path,
        inet_files_rel,
    )

    offline_rustup_path = offline_path / "rustup"
    offline_args = [
        "--archive",
        f"{archive_path}",
    ]

    rustup_must_run(
        offline_path,
        offline_args + ["unpack"],
    )

    offline_files_rel = set(walk_files_rel(offline_rustup_path))

    assert_same_files(
        inet_rustup_path,
        inet_files_rel,
        offline_rustup_path,
        offline_files_rel,
    )



import datetime
import os
import platform
import re
import stat
import tarfile
from contextlib import contextmanager
from pathlib import Path
from typing import (
    IO,
    Any,
    Generator,
    Iterable,
    List,
    NoReturn,
    Optional,
    Tuple,
)

import romt.error

is_windows = platform.system() == "Windows"


VERBOSITY_ERROR = 0
VERBOSITY_INFO = 1
VERBOSITY_VERBOSE = 2
VERBOSITY_VVERBOSE = 3
_max_verbosity = VERBOSITY_INFO


def get_max_verbosity() -> int:
    return _max_verbosity


def set_max_verbosity(max_verbosity: int) -> None:
    global _max_verbosity
    _max_verbosity = max_verbosity


def _print_verbosity(verbosity: int, *args: Any, **kwargs: Any) -> None:
    if verbosity <= _max_verbosity:
        print(*args, flush=True, **kwargs)


def vvprint(*args: Any, **kwargs: Any) -> None:
    _print_verbosity(VERBOSITY_VVERBOSE, *args, **kwargs)


def vprint(*args: Any, **kwargs: Any) -> None:
    _print_verbosity(VERBOSITY_VERBOSE, *args, **kwargs)


def iprint(*args: Any, **kwargs: Any) -> None:
    _print_verbosity(VERBOSITY_INFO, *args, **kwargs)


def eprint(*args: Any, **kwargs: Any) -> None:
    _print_verbosity(VERBOSITY_ERROR, *args, **kwargs)


def abort(*args: Any, **kwargs: Any) -> NoReturn:
    eprint(*args, **kwargs)
    raise romt.error.AbortError()


def is_date(date: str) -> bool:
    return re.match(r"\d\d\d\d-\d\d-\d\d$", date) is not None


def is_version(version: str) -> bool:
    return re.match(r"\d+\.\d+\.\d+$", version) is not None


def version_sort_key(version: str) -> Tuple[int, ...]:
    return tuple(int(v) for v in version.split("."))


def reverse_sorted_versions(versions: List[str]) -> List[str]:
    return sorted(versions, key=version_sort_key, reverse=True)


def path_append(path: Path, suffix: str) -> Path:
    return path.with_name(path.name + suffix)


def tmp_path_for(path: Path) -> Path:
    return path.with_name("." + path.name + ".tmp")


def path_modified_today(path: Path) -> bool:
    if path.is_file():
        path_mtime = path.stat().st_mtime
        path_mdate = datetime.datetime.fromtimestamp(path_mtime).date()
        today_date = datetime.datetime.now().date()
        modified_today = path_mdate == today_date
    else:
        modified_today = False
    return modified_today


def get_umask() -> int:
    umask = os.umask(0)
    os.umask(umask)
    return umask


def is_executable(path: Path) -> bool:
    return os.access(str(path), os.X_OK)


def chmod_executable(path: Path) -> None:
    x_bits = (stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH) & ~get_umask()
    os.chmod(str(path), path.stat().st_mode | x_bits)


def gen_dirs(parent: Path) -> Generator[Path, None, None]:
    """generate Path for each dir in parent."""
    for candidate in parent.glob("*"):
        if candidate.is_dir():
            yield candidate


def reversed_date_dir_names(parent: Path) -> List[str]:
    """list of yyyy-mm-dd dirnames in parent (newest to oldest)."""
    dirs = sorted(
        (d.name for d in gen_dirs(parent) if is_date(d.name)),
        reverse=True,
    )
    return dirs


def open_optional(path: str, mode: str) -> Optional[IO[Any]]:
    return open(path, mode) if path else None


def close_optional(f: Optional[IO[Any]]) -> None:
    if f:
        f.close()


def make_dirs_for(path: Path) -> None:
    parent = path.parent
    if not parent.is_dir():
        parent.mkdir(parents=True)


def log(log_file: Optional[IO[Any]], message: Any) -> None:
    if log_file is not None:
        log_file.write(f"{message}\n")
        log_file.flush()


def split_word(item: str) -> List[str]:
    """split item into list of words at commas or runs of whitespace.

    Retains any duplicates and empty strings.
    """
    return [part for part in re.split(r"(?:,|\s+)", item)]


def split_flatten_words(words: Iterable[str]) -> List[str]:
    """split_word(each_word in words) into flattened, deduped list."""
    dedup = set()
    result = []
    for w in words:
        for part in split_word(w):
            if part not in dedup:
                dedup.add(part)
                result.append(part)
    return result


def split_flatten_normalize_words(words: Iterable[str]) -> List[str]:
    """split_flatten_words(), remove dups and empty, sort."""
    norm_words = {w for w in split_flatten_words(words) if w}
    return sorted(norm_words)


def normalize_patterns(patterns: Iterable[str]) -> List[str]:
    """split, flatten, remove dups and empty, reduce "*", sort."""
    norm_patterns = split_flatten_normalize_words(patterns)
    if "*" in norm_patterns:
        return ["*"]
    return norm_patterns


@contextmanager
def tar_context(
    archive_path: Path, mode: str
) -> Generator[tarfile.TarFile, None, None]:
    """mode is "r" (read) or "w" (write)."""
    writing = mode == "w"
    if archive_path.name.endswith(".gz"):
        mode += ":gz"

    if writing:
        if archive_path.exists():
            archive_path.unlink()
        tmp_archive_path = tmp_path_for(archive_path)
        tar_f = tarfile.open(str(tmp_archive_path), mode)
        try:
            yield tar_f
        except (Exception, KeyboardInterrupt):
            tar_f.close()
            if tmp_archive_path.is_file():
                tmp_archive_path.unlink()
            raise
        tar_f.close()
        tmp_archive_path.rename(archive_path)
    else:
        tar_f = tarfile.open(str(archive_path), mode)
        tar_f.extraction_filter = tarfile.data_filter
        yield tar_f
        tar_f.close()



import collections
import copy
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Iterable,
    List,
    MutableMapping,
    Optional,
    Tuple,
)

import toml


def target_matches_any(target: str, expected_targets: Iterable[str]) -> bool:
    if target == "*":
        return True
    for expected in expected_targets:
        if target == expected or expected == "*":
            return True
    return False


class Package:
    def __init__(
        self, name: str, target: str, details: MutableMapping[str, Any]
    ):
        self.name = name
        self.target = target
        self.available = details["available"]
        self.xz_url = str(details.get("xz_url", ""))

    def _fields(self) -> Tuple[str, str]:
        return (self.name, self.target)

    def __repr__(self) -> str:
        return f"Package {{ name={self.name}, target={self.target} }}"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Package):
            return False
        return self._fields() == other._fields()

    def __hash__(self) -> int:
        return hash(self._fields())

    @property
    def has_rel_path(self) -> bool:
        return bool(self.xz_url != "")

    @property
    def rel_path(self) -> str:
        if not self.has_rel_path:
            raise ValueError(
                f"Package {self.name}/{self.target} missing xz_url"
            )
        url = self.xz_url
        prefix = "/dist/"
        return url[url.index(prefix) + len(prefix) :]


class Manifest:
    def __init__(self, raw_dict: MutableMapping[str, Any]):
        self._dict = raw_dict

    @staticmethod
    def from_toml_path(toml_path: Path) -> "Manifest":
        return Manifest(toml.load(toml_path))

    def clone(self) -> "Manifest":
        return Manifest(copy.deepcopy(self._dict))

    @property
    def _rust_src_version(self) -> str:
        version = str(self._dict["pkg"]["rust-src"]["version"])
        # Sample version lines found below [pkg.rust-src]:
        # version = "1.43.0-beta.5 (934ae7739 2020-04-06)"
        # version = "1.44.0-nightly (42abbd887 2020-04-07)"
        # version = "1.42.0 (b8cedc004 2020-03-09)"
        return version

    @property
    def channel(self) -> str:
        version = self._rust_src_version
        if "-beta" in version:
            channel = "beta"
        elif "-nightly" in version:
            channel = "nightly"
        else:
            channel = "stable"
        return channel

    @property
    def version(self) -> str:
        version = self._rust_src_version
        # version = "1.44.0-nightly (42abbd887 2020-04-07)"
        # version = "1.42.0 (b8cedc004 2020-03-09)"
        return version.split("-")[0].split()[0]

    @property
    def date(self) -> str:
        return str(self._dict["date"])

    @property
    def spec(self) -> str:
        return f"{self.channel}-{self.date}"

    @property
    def ident(self) -> str:
        return f"{self.spec}({self.version})"

    def get_package(self, package_name: str, target: str) -> Package:
        details = self._dict["pkg"][package_name]["target"][target]
        return Package(package_name, target, details)

    def gen_all_packages(self) -> Generator[Package, None, None]:
        """Generate Package for all (name, target) in manifest."""
        for name, package_dict in self._dict["pkg"].items():
            for target in package_dict["target"].keys():
                yield self.get_package(name, target)

    def gen_available_packages(
        self,
        *,
        targets: Optional[Iterable[str]] = None,
        rel_path_is_present: Optional[Callable[[str], bool]] = None,
    ) -> Generator[Package, None, None]:
        """available packages matching targets and "present"."""
        if targets is None:
            target_list = ["*"]
        else:
            target_list = list(targets)
        for package in self.gen_all_packages():
            if (
                package.available
                and target_matches_any(package.target, target_list)
                and (
                    rel_path_is_present is None
                    or rel_path_is_present(package.rel_path)
                )
            ):
                yield package

    def all_targets(self) -> List[str]:
        targets = {p.target for p in self.gen_all_packages()}
        targets.discard("*")
        return sorted(targets)

    def available_targets(
        self,
        *,
        targets: Optional[Iterable[str]] = None,
        rel_path_is_present: Optional[Callable[[str], bool]] = None,
    ) -> List[str]:
        available_targets = {
            p.target
            for p in self.gen_available_packages(
                targets=targets,
                rel_path_is_present=rel_path_is_present,
            )
        }
        available_targets.discard("*")
        return sorted(available_targets)

    def available_target_types(
        self,
        *,
        targets: Optional[Iterable[str]] = None,
        rel_path_is_present: Optional[Callable[[str], bool]] = None,
    ) -> Dict[str, str]:
        target_packages = collections.defaultdict(set)
        rel_path_targets = collections.defaultdict(set)
        for package in self.gen_available_packages():
            # No need to discard the case `target == "*"`.
            target_packages[package.target].add(package)
            rel_path_targets[package.rel_path].add(package.target)

        target_types: Dict[str, str] = {}
        if targets is not None:
            target_list = list(targets)
        else:
            target_list = self.available_targets()
        for target in sorted(target_list):
            packages = target_packages[target]
            if not packages:
                continue
            have_all_rel_paths = True
            have_unique_rel_path = False
            have_rustc = False
            have_rust_std = False
            for package in packages:
                is_present = (
                    rel_path_is_present is None
                    or rel_path_is_present(package.rel_path)
                )
                if is_present:
                    if package.name == "rustc":
                        have_rustc = True
                    elif package.name == "rust-std":
                        have_rust_std = True
                    if len(rel_path_targets[package.rel_path]) == 1:
                        have_unique_rel_path = True
                else:
                    have_all_rel_paths = False
            if have_unique_rel_path or have_all_rel_paths:
                if have_rustc:
                    target_type = "native-target"
                elif have_rust_std:
                    target_type = "cross-target"
                else:
                    target_type = "minimal"
                target_types[target] = target_type

        return target_types
